StellarBars = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceHook-2.1", "AceDB-2.0")
local StellarBars = StellarBars
local self = StellarBars

local dewdrop = AceLibrary("Dewdrop-2.0")
local smed = AceLibrary("SharedMedia-1.0")
smed:Register("statusbar", "Waves", "Interface\\AddOns\\StellarBars\\waves")

-- GLOBAL -> LOCAL
local _G = getfenv(0)
local main = MainMenuBar
local multibl, multibr, multil, multir = MultiBarBottomLeft, MultiBarBottomRight, MultiBarLeft, MultiBarRight
local mainexp, repexp, ReputationWatchStatusBar = MainMenuExpBar, ReputationWatchBar, ReputationWatchStatusBar
local shiftbar, petbar = ShapeshiftBarFrame, PetActionBarFrame
local GetNumShapeshiftForms = GetNumShapeshiftForms
local InCombatLockdown = InCombatLockdown

-- LOCAL VARS
local db, opts, done
local buttons = {
	"ActionButton", "BonusActionButton",
	"MultiBarBottomLeftButton", "MultiBarBottomRightButton",
	"MultiBarLeftButton", "MultiBarRightButton",
	"ShapeshiftButton", "PetActionButton",
}

-- LOCAL FUNCS
-- complete setpoint adjustments
local function FullAdjust(frame, p, rt, rp, x, y, parent)
	local a,b,c,d,e = frame:GetPoint()
	frame:ClearAllPoints()
	frame:SetPoint(p or a, rt or b, rp or c, x or d, y or e)
end
-- x/y setpoint adjustments
local function XYAdjust(frame, x, y)
	local a,b,c,d,e = frame:GetPoint()
	frame:SetPoint(a,b,c, x or d, y or e)
end
local ShowOptions

-----------------------------------
function StellarBars:OnInitialize()
-----------------------------------
	self:RegisterDB("StellarBarsDB")
	self:RegisterDefaults("profile", {
		scale = 0.9,
		mainenable = true,
		mainx = 0,
		mainy = 0,
		sideenable = true,
		sidex = 0,
		sidey = 0,
		r = 0,
		g = 0,
		b = 0,
		texture = "Waves",
		oor = true,
		selfcast = true,
	} )
	db = self.db.profile
	
	SlashCmdList.STELLARBARS = ShowOptions
	SLASH_STELLARBARS1, SLASH_STELLARBARS2 = "/stellarbars", "/stellarbar"
	self:HookScript(MainMenuBarPerformanceBarFrameButton, "OnClick", ShowOptions)
	
	self:ScheduleLeaveCombatAction("ApplyTheLook")
end

-----------------------------------
function StellarBars:ApplyTheLook()
-----------------------------------
	self:HideOldStuff()
	self:ShowNewStuff()
	
	-- refresh position whenever the UIParent does
	if AceLibrary:HasInstance("Jostle-2.0") then
		-- Jostle work around
		self:SecureHook(AceLibrary("Jostle-2.0"), "Refresh", function(_,f)
			if f == nil or f == GroupLootFrame1 or f == main or f == multir then
				if InCombatLockdown() then
					if not done then
						done = true
						self:ScheduleLeaveCombatAction("RefreshAdjustments")
					end
					return
				end
				self:RefreshAdjustments()
			end
		end)
		self:Hook("UIParent_ManageFramePositions", function()
			if InCombatLockdown() then return end
			self.hooks.UIParent_ManageFramePositions()
		end, true)
	else
		self:Hook("UIParent_ManageFramePositions", function()
			if InCombatLockdown() then
				if not done then
					done = true
					self:ScheduleLeaveCombatAction(UIParent_ManageFramePositions)
				end
				return
			end
			self.hooks.UIParent_ManageFramePositions()
			self:RefreshAdjustments()
		end, true)
	end
	
	-- hack to make the petbar show up in combat that many mods' tainting prevents (maybe this one too :p)
	petbar:SetAttribute("unit", "pet")
	RegisterUnitWatch(petbar)
	self:Hook("ShowPetActionBar", function() UIParent_ManageFramePositions() end, true)
	self:Hook("HidePetActionBar", function() end, true)
	
	self:SecureHook("PetActionBar_Update", function()
		for i=1, NUM_PET_ACTION_SLOTS, 1 do
			_G["PetActionButton"..i.."NormalTexture2"]:Hide()
		end
	end)
	
	-- adds right-click self-casting
	if db.selfcast then
		MainMenuBarArtFrame:SetAttribute("unit2", "player")
		multibl:SetAttribute("unit2", "player")
		multibr:SetAttribute("unit2", "player")
		multil:SetAttribute("unit2", "player")
		multir:SetAttribute("unit2", "player")
	end
	if db.oor then
		self:SecureHook("ActionButton_OnUpdate")
		self:SecureHook("ActionButton_UpdateUsable")
	end
	
	self:RefreshAdjustments()
end

-----------------------------------
function StellarBars:HideOldStuff()
-----------------------------------
	-- hide some texture elements
	local stuffToHide = {
		"MainMenuXPBarTexture0", "MainMenuXPBarTexture1", "MainMenuXPBarTexture2", "MainMenuXPBarTexture3",
		"ReputationWatchBarTexture0", "ReputationWatchBarTexture1", "ReputationWatchBarTexture2", "ReputationWatchBarTexture3",
		"ReputationXPBarTexture0", "ReputationXPBarTexture1", "ReputationXPBarTexture2", "ReputationXPBarTexture3",
		"MainMenuBarTexture0", "MainMenuBarTexture1", "MainMenuBarTexture2", "MainMenuBarTexture3",
		"MainMenuBarLeftEndCap", "MainMenuBarRightEndCap",
		"ShapeshiftBarLeft", "ShapeshiftBarRight", "ShapeshiftBarMiddle",
		"MainMenuMaxLevelBar0", "MainMenuMaxLevelBar1", "MainMenuMaxLevelBar2", "MainMenuMaxLevelBar3",
		"BonusActionBarTexture0", "BonusActionBarTexture1",
		"SlidingActionBarTexture0", "SlidingActionBarTexture1",
	}
	for index,value in pairs(stuffToHide) do
		_G[value]:SetAlpha(0)
		_G[value]:Hide()
	end
	
	-- hide borders for action, bonus, pet, stance, and extra buttons
	for _,value in pairs(buttons) do
		for i=1,12 do
			local butt = format("%s%d", value, i)
			if _G[butt] then
				_G[butt.."Icon"]:SetTexCoord(0.07, 0.93, 0.07, 0.93)
				if value == "PetActionButton" then
					_G[butt.."NormalTexture2"]:Hide()
				else
					_G[butt.."NormalTexture"]:Hide()
				end
			end
		end
	end
	
	-- hide borders for bag icons
	MainMenuBarBackpackButtonIconTexture:SetTexCoord(0.07, 0.93, 0.07, 0.93)
	MainMenuBarBackpackButtonNormalTexture:Hide()
	for i=0,3 do
		_G["CharacterBag"..i.."SlotIconTexture"]:SetTexCoord(0.07, 0.93, 0.07, 0.93)
		_G["CharacterBag"..i.."SlotNormalTexture"]:Hide()
	end
end


-- backgrounds
-- creates the background textures
do
	local backdrop1 = { bgFile = "Interface\\AddOns\\StellarBars\\backdrop", }
	local backdrop2 = { bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 8, }
	-- creates a new background for a button
	local function DoCreate(frame, r, g, b, p)
		local f = _G[frame]
		f:SetFrameStrata("MEDIUM")
		local fbg = CreateFrame("Frame", frame.."BG", (p and main) or f)
		fbg:SetFrameStrata("LOW")
		fbg:SetPoint("TOPLEFT", f, "TOPLEFT", -2, 2)
		fbg:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", 2, -2)
		fbg:SetBackdrop(backdrop1)
		fbg:SetBackdropColor(r, g, b, 0.9)
		fbg:EnableMouse(true)
	end
	local function alterexpbar(f, ftext)
		f:SetHeight(8)
		f:SetWidth(f:GetWidth()-16)
		
		local ft1 = f:CreateTexture(nil, "OVERLAY")
		ft1:SetTexture("Interface\\AddOns\\StellarBars\\ticks")
		ft1:SetPoint("TOPLEFT", f, "TOPLEFT", 0, 0)
		ft1:SetPoint("BOTTOMRIGHT", f, "BOTTOM", 0, 0)
		ft1:SetVertexColor(0.9,0.9,0.9,0.9)
		
		local ft2 = f:CreateTexture(nil, "OVERLAY")
		ft2:SetTexture("Interface\\AddOns\\StellarBars\\ticks")
		ft2:SetPoint("TOPLEFT", f, "TOP", 0, 0)
		ft2:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", 0, 0)
		ft2:SetVertexColor(0.9,0.9,0.9,0.9)
		
		ftext:SetFont("Fonts\\FRIZQT__.TTF", 11.5)
		ftext:SetVertexColor(1.0, 1.0, 1.0)
		ftext:SetShadowOffset(-1.5,-1.5)
	end

	-----------------------------------
	function StellarBars:ShowNewStuff()
	-----------------------------------
		-- make backgrounds for action buttons
		local r, g, b = db.r, db.g, db.b
		local notwar = select(2, UnitClass("player")) ~= "WARRIOR"
		for _,value in pairs(buttons) do
			if value ~= "ActionButton" or notwar then
				for i=1,12 do
					local butt = format("%s%d", value, i)
					local butthk = _G[format("%sHotKey", butt)]
					if _G[butt] then
						DoCreate(butt, r, g, b, value == "ActionButton" or value == "BonusActionButton")
					end
					if butthk then
						butthk:SetFont("Fonts\\FRIZQT__.TTF", 12)
						butthk:SetShadowOffset(-1.5,-1.5)
					end
				end
			end
		end
		DoCreate("MainMenuBarBackpackButton", r, g, b)
		for i=0,3 do
			DoCreate("CharacterBag"..i.."Slot", r, g, b)
		end
		
		-- bonus action buttons
		for i = 1,12,1 do
			_G["ActionButton"..i]:SetFrameStrata("MEDIUM")
			_G["BonusActionButton"..i]:SetFrameStrata("DIALOG")
			_G["BonusActionButton"..i.."BG"]:SetFrameStrata("HIGH")
			_G["BonusActionButton"..i.."BG"]:SetBackdropColor(r, g, b, 0.99)
		end
		self:SecureHook("ShowBonusActionBar", function()
			for i = 1,12,1 do
				_G["BonusActionButton"..i.."BG"]:Show()
			end
		end)
		self:SecureHook("HideBonusActionBar", function()
			for i = 1,12,1 do
				_G["BonusActionButton"..i.."BG"]:Hide()
			end
		end)
		if not BonusActionBarFrame:IsShown() then
			for i = 1,12,1 do
				_G["BonusActionButton"..i.."BG"]:Hide()
			end
		end	
		
		-- adds a background behind the etc buttons
		local newBG = CreateFrame("Button", "MainEtcBG", main)
		newBG:SetPoint("TOPLEFT", ActionBarUpButton, "TOPLEFT", 4, -5)
		newBG:SetPoint("BOTTOMRIGHT", KeyRingButton, "BOTTOMRIGHT", 1, 1)
		newBG:SetBackdrop(backdrop2)
		newBG:SetBackdropColor(r, g, b, 0.75)
		newBG:SetFrameStrata("LOW")
		newBG:SetFrameLevel(1)
		
		-- adjust experience bars
		alterexpbar(mainexp, MainMenuBarExpText)
		mainexp:HookScript("OnEnter", function()
			local this = this
			local vmin, vmax = this:GetMinMaxValues()
			this.TextString:SetText(format("%s (%0.1f%%)", this.TextString:GetText(), this:GetValue()*100/(vmax-vmin)))
		end)
		alterexpbar(ReputationWatchStatusBar, ReputationWatchStatusBarText)
		repexp:SetWidth(repexp:GetWidth()-16)
		self:SecureHook("ReputationWatchBar_Update", function() 
			local this = ReputationWatchStatusBar
			local vmin, vmax = this:GetMinMaxValues()
			ReputationWatchStatusBarText:SetText(format("%s (%0.1f%%)", ReputationWatchStatusBarText:GetText() or "", this:GetValue()*100/(vmax-vmin)))
			this:SetHeight(8) 
			FullAdjust(repexp, "TOP", (mainexp:IsShown() and mainexp) or main, "BOTTOM", nil, -1)
		end)

		-- adjust action buttons
		FullAdjust(mainexp, "TOP", nil, "BOTTOM")
		FullAdjust(repexp, "TOP", mainexp, "BOTTOM")
		FullAdjust(multibl, "BOTTOMLEFT", main, "TOPLEFT", 8, -8)
		FullAdjust(multibr, "BOTTOMRIGHT", main, "TOPRIGHT", -4, -8)
		FullAdjust(main, nil, nil, nil, nil, 20)
		shiftbar:SetScale(0.92)
		petbar:SetScale(0.92)
		self:UpdateLook()
	end
end

-- layout tweaking, called often
-----------------------------------------
function StellarBars:RefreshAdjustments()
-----------------------------------------
	if db.mainenable then
		XYAdjust(multir, ((db.sidex == 0 and -7) or db.sidex)/0.92, (98 + db.sidey)/db.scale)
	end
	if db.sideenable then
		XYAdjust(main, db.mainx/db.scale, (db.mainy ~= 0 and db.mainy/db.scale) or nil)
	end

	multibl:SetPoint("BOTTOMLEFT", main, "TOPLEFT", 8, -8)
	multibr:SetPoint("BOTTOMRIGHT", main, "TOPRIGHT", -4, -8)
	shiftbar:SetPoint("BOTTOMLEFT", main, "TOPLEFT", 30, ((multibl:IsShown() and 30) or -12)/0.92)
	
	local xpet = (GetNumShapeshiftForms() > 0 and 360) or 30
	if multibl:IsShown() or (shiftbar:IsShown() and multibr:IsShown()) then
		petbar:SetPoint("TOPLEFT", main, "BOTTOMLEFT", xpet, 123/0.92)
	else
		petbar:SetPoint("TOPLEFT", main, "BOTTOMLEFT", xpet, 81/0.92)
	end
	
	FullAdjust(repexp, "TOP", (mainexp:IsShown() and mainexp) or main, "BOTTOM", nil, -1)
	done = nil
end


-- oor and oom recoloring
do
	local getpage, isusable, isrange = ActionButton_GetPagedID, IsUsableAction, IsActionInRange
	local elap = { }
	---------------------------------------------------
	function StellarBars:ActionButton_OnUpdate(elapsed)
	---------------------------------------------------
		local n = this:GetName()
		_G[n.."HotKey"]:SetVertexColor(0.95, 0.95, 0.95)
		elap[n] = (elap[n] or 0) + elapsed
		if elap[n] < 0.25 then return end
		elap[n] = 0
		self:ActionButton_UpdateUsable(n)
	end
	----------------------------------------------------
	function StellarBars:ActionButton_UpdateUsable(name)
	----------------------------------------------------
		local n = name or this:GetName()
		local page = getpage(this)
		local isUsable, notEnoughMana = isusable(page)
		if isUsable then
			if isrange(page) ~= 0 then
				_G[n.."Icon"]:SetVertexColor(1.0, 1.0, 1.0)
			else
				_G[n.."Icon"]:SetVertexColor(1.0, 0.5, 0.5)
			end
		elseif notEnoughMana then
			_G[n.."Icon"]:SetVertexColor(0.5, 0.5, 1.0)
		else
			_G[n.."Icon"]:SetVertexColor(0.4, 0.4, 0.4)
		end
		_G[n.."NormalTexture"]:Hide()
	end
end

-----------------------------------
function StellarBars:UpdateLook()
-----------------------------------
	local s = db.scale
	main:SetScale(s)
	multibl:SetScale(s)
	multibr:SetScale(s)
	multir:SetScale(s)
	multil:SetScale(s)
	
	local t = smed:Fetch("statusbar", db.texture)
	mainexp:SetStatusBarTexture(t)
	ReputationWatchStatusBar:SetStatusBarTexture(t)
end

------------------------------------------
function StellarBars:UpdateColors(r, g, b)
------------------------------------------
	for _,value in pairs(buttons) do
		for i=1,12 do
			if _G[value..i.."BG"] then
				_G[value..i.."BG"]:SetBackdropColor(r, g, b, 0.9)
			end
		end
	end
	
	MainMenuBarBackpackButtonBG:SetBackdropColor(r, g, b, 0.9)
	for i=0,3 do
		_G["CharacterBag"..i.."SlotBG"]:SetBackdropColor(r, g, b, 0.9)
	end
	MainEtcBG:SetBackdropColor(r, g, b, 0.75)
end

-- Options table
do
	local function set(k, v, v2, v3)
		if k == "color" then
			db.r, db.g, db.b = v, v2, v3
			self:UpdateColors(v, v2, v3)
		else
			db[k] = v
			self:RefreshAdjustments()
			self:UpdateLook()
		end
	end
	local function get(k)
		if k == "color" then
			return db.r, db.g, db.b
		else
			return db[k]
		end
	end
	local reloaddesc = "Note: May require ReloadUI to take full effect"
	opts = {
		type = "group",
		args = {
			header = { type = "header", name = "StellarBars", order = 1, },
			main = {
				type = "group", name = "Mainbar", desc = "Adjust position of the mainbar",
				pass = true, set = set, get = get,
				order = 3,
				args = {
					mainenable = { 
						type = "toggle", name = "Enable adjustments", desc = reloaddesc, 
						order = 1, 
					},
					mainx = {
						type = "range", name = "X", desc = "Horizontal adjustments",
						min = -500, max = 500, step = 1, bigStep = 5,
						disabled = InCombatLockdown,
						order = 2,
					},
					mainy = {
						type = "range", name = "Y", desc = "0 for disabled. "..reloaddesc,
						min = -40, max = 1200, step = 1, bigStep = 5,
						disabled = InCombatLockdown,
						order = 3
					},
					scale = {
						type = "range", name = "Scale", desc = "Change scale",
						min = 0.5, max = 1.5, step = 0.02,
						disabled = InCombatLockdown,
						order = 4,
					},
					texture = {
						type = "text", name = "Experience bar texture", desc = "Experience bar texture",
						validate = smed:List("statusbar"),
						order = 5,
					},
				},
			},
			side = {
				type = "group", name = "Sidebar", desc = "Adjust position of the sidebar",
				pass = true, set = set, get = get,
				order = 4,
				args = {
					sideenable = { 
						type = "toggle", name = "Enable adjustments", desc = reloaddesc, 
						order = 1, 
					},
					sidex = {
						type = "range", name = "X", desc = "0 for disabled. "..reloaddesc,
						min = -2000, max = 100, step = 1, bigStep = 10,
						disabled = InCombatLockdown,
						order = 2,
					},
					sidey = {
						type = "range", name = "Y", desc = "Vertical adjustments",
						min = -500, max = 1000, step = 1, bigStep = 5,
						disabled = InCombatLockdown,
						order = 3
					},
				},
			},
			color = { 
				type = "color", name = "Color", desc = "Change the color of the layout", 
				set = set, get = get, passValue = "color",
				order = 5, 
			},
			oor = { 
				type = "toggle", name = "Out-of-range coloring", desc = reloaddesc, 
				set = set, get = get, passValue = "oor",
				order = 6, 
			},
			selfcast = { 
				type = "toggle", name = "Right-click self-cast", desc = reloaddesc, 
				set = set, get = get, passValue = "selfcast",
				order = 7, 
			},
		},
	}
	function ShowOptions() 
		dewdrop:Open( UIParent, 'children', function() dewdrop:FeedAceOptionsTable(opts) end, 'cursorX', true, 'cursorY', true )
	end

end